//
// MoonCanvas.cpp
// Draw the moon and related info.
//
//    Copyright 2002 by Akkana Peck.
// You may use or distribute this code under the terms of the GPL.
//

#include "MoonCanvas.h"

#include "MoonFeature.h"

#include <qpainter.h>
#include <qimage.h>
#include <qmessagebox.h>

// XXX next two should be removable:
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define DEGMAX 100
#define DEGMIN  -5

int MoonCanvas::sDebugMode = 0;

MoonCanvas::MoonCanvas( QWidget *parent, const char *name )
    : QWidget( parent, name ),
      mMoonRadius(0), mMoonCenterX(-1), mMoonCenterY(-1),
      mMoonImg(0),
      mMirrored(1), mZoomLevel(0),
      mPainter(0), mCraterNum(0)
{
}

MoonCanvas::~MoonCanvas()
{
    if (mPainter)
        delete mPainter;
    if (mMoonImg)
        delete mMoonImg;
}

void MoonCanvas::Clear()
{
    if (sDebugMode)
        printf("MoonCanvas::Clear\n");
    if (!mPainter)
        return;
    mPainter->setBackgroundColor(QColor("black"));
    mPainter->eraseRect(0, 0, width(), height());
}

void MoonCanvas::Zoom(int level)
{
    if (level > 1) level = 1;
    if (level < 0) level = 0;
    mZoomLevel = level;

    if (mMoonImg)
        delete mMoonImg;
    mMoonImg = 0;

    if (mZoomLevel > 0)
    {
        QString dir("/opt/Qtopia/share/qmoon/");
        QString mirror("-mirror");
        QString nomirror("");
        QString extension(".png");
        QString quadstr;
        int quad;
        // Figure out where we are from our object list:
        if (mCraterNum <= 0)
            quad = 1;
        else
        {
            int longitude = mMatches[0]->GetLongitude();
            int latitude = mMatches[0]->GetLatitude();
            if (longitude < 0 && latitude > 0)
                quad = 1;
            else if (longitude > 0 && latitude > 0)
                quad = 2;
            else if (longitude < 0 && latitude < 0)
                quad = 3;
            else 
                quad = 4;
        }

        // Open the image:
        QString moonImgStr = dir + "quad" + QString::number(quad)
                              + (mMirrored < 0 ? mirror : nomirror)
                              + extension;
        mMoonImg = new QImage(moonImgStr);
        if (sDebugMode)
            printf("Opened %s: %d x %d\n",
                   (const char*)moonImgStr,
                   mMoonImg->width(), mMoonImg->height());
        mMoonRadius = (int)(.972 * mMoonImg->width());
        mImgLibrationE = 4;
        mImgLibrationN = 1;

        // Fine-tune the offsets for these images (what a pain):
        switch (quad)
        {
          case 2:
              mMinLong = DEGMIN;
              mMaxLong = DEGMAX;
              mMinLat = DEGMIN;
              mMaxLat = DEGMAX;
              quadstr = "quad2";
              // The moon center numbers are specific to these images.
              mMoonCenterY = (int)(.976*mMoonImg->height());
              if (mMirrored < 0)
                  mMoonCenterX = (int)(.963*mMoonImg->width());
              else
                  mMoonCenterX = (int)(.040*mMoonImg->width());
              break;
          case 3:
              mMinLong = -DEGMAX;
              mMaxLong = -DEGMIN;
              mMinLat = -DEGMIN;
              mMaxLat = -DEGMAX;
              quadstr = "quad3";
              mMoonCenterY = (int)(.046*mMoonImg->height());
              if (mMirrored < 0)
                  mMoonCenterX = (int)(.016*mMoonImg->width());
              else
                  mMoonCenterX = (int)(.986*mMoonImg->width());
              break;
          case 4:
              mMinLong = DEGMAX;
              mMaxLong = DEGMIN;
              mMinLat = -DEGMIN;
              mMaxLat = -DEGMAX;
              quadstr = "quad4";
              mMoonCenterY = (int)(.033*mMoonImg->height());
              if (mMirrored < 0)
                  mMoonCenterX = (int)(.958*mMoonImg->width());
              else
                  mMoonCenterX = (int)(.041*mMoonImg->width());
              break;
          case 0:
          case 1:
          default:
              mMinLong = -DEGMAX;
              mMaxLong = -DEGMIN;
              mMinLat = DEGMIN;
              mMaxLat = DEGMAX;
              quadstr = "quad1";
              mMoonCenterY = (int)(.980*mMoonImg->height());
              if (mMirrored < 0)
                  mMoonCenterX = (int)(.023*mMoonImg->width());
              else
                  mMoonCenterX = (int)(.976*mMoonImg->width());
              break;
        }

        if (mMoonRadius > 1)
            return;

        // If we fell through to here, there's something wrong with
        // the image.  So we'll just fall back to zoom level 1.
        if (mMoonImg)
            delete mMoonImg;
    }

    // Zoom level 1 (full moon):
    mMinLong = mMinLat = -DEGMAX;
    mMaxLong = mMaxLat = DEGMAX;
    mImgLibrationE = 5;
    mImgLibrationN = 0;
    if (mMirrored < 0) {
        mMoonImg = new QImage
            (QString("/opt/Qtopia/share/qmoon/fullmoon-mirror.png"));
        if (sDebugMode)
            printf("Opened fullmoon-mirror.png: %d x %d\n",
                   mMoonImg->width(), mMoonImg->height());
    } else {
        mMoonImg = new QImage
            (QString("/opt/Qtopia/share/qmoon/fullmoon.png"));
        if (sDebugMode)
            printf("Opened fullmoon.png: %d x %d\n",
                   mMoonImg->width(), mMoonImg->height());
    }

    setMinimumWidth(mMoonImg->width());
    setMinimumHeight(mMoonImg->height());
    if (sDebugMode)
        printf("Setting min size to %d x %d\n",
               mMoonImg->width(), mMoonImg->height());

    mMoonCenterX = mMoonImg->width()/2;
    mMoonCenterY = mMoonImg->height()/2;
    if (sDebugMode)
        printf("Center at (%d, %d)\n", mMoonCenterX, mMoonCenterY);

    mMoonRadius = mMoonImg->width()/2 - 2;
}

void MoonCanvas::Init()
{
    Zoom(0);

    //mPainter = new QPainter(this);

    MoonFeature::InitMoonFeatures(mMoonRadius);

    MoonFeature::SetCanvasSize(width(), height());
}

void MoonCanvas::ToggleMirrored()
{
    mMirrored = -mMirrored;
    Zoom(mZoomLevel);

    Paint();
}

#if 0
QSize MoonCanvas::minimumSizeHint() const
{
    return QSize(width(), width());
}
#endif

void MoonCanvas::paintEvent(QPaintEvent*)
{
    //printf("MoonCanvas::paintEvent\n");
    Paint();
}

void MoonCanvas::Paint()
{
    //printf("MoonCanvas::Paint\n");
    if (mMoonRadius <= 0)
        Init();

    if (!mPainter)
        mPainter = new QPainter(this);

    Clear();
    mPainter->setRasterOp(Qt::CopyROP);
    mPainter->drawImage(0, 0, *mMoonImg);
    //mPainter->drawImage(QPoint(0, 0), *mMoonImg);
    int i;
    for (i=0; i < mCraterNum; ++i)
        mMatches[i]->PaintFeature(mPainter, mMoonCenterX, mMoonCenterY,
                                  mMoonRadius,
                                  mMirrored, mImgLibrationE, mImgLibrationN);

    delete mPainter;
    mPainter = 0;

    return;
}

void MoonCanvas::PaintFeature(MoonFeature* feature)
{
    if (!mPainter)
        mPainter = new QPainter(this);
    feature->PaintFeature(mPainter, mMoonCenterX, mMoonCenterY, mMoonRadius,
                          mMirrored, mImgLibrationE, mImgLibrationN);

    mDescriptions.append(feature->GetName());
    QString numstr;
    numstr.setNum(feature->GetSize());
    mDescriptions.append(" (");
    mDescriptions.append(numstr);
    mDescriptions.append(" km)\n");
    mDescriptions.append(feature->GetDesc());
    mDescriptions.append("\n");
}

void MoonCanvas::mousePressEvent(QMouseEvent *event)
{
    // Remember the time:
    gettimeofday(&mLastButtonDownTime, 0);

    // We're only interested in doubleclick events:
    if (event->type() != QEvent::MouseButtonDblClick) {
        mCraterNum = 0;
        mDescriptions.truncate(0);
    }
    else
        QMessageBox::information(this, "QMoon", mDescriptions,
                                 QMessageBox::Ok);
}

void MoonCanvas::mouseReleaseEvent(QMouseEvent *event)
{
    // Get the time of the event -- we'll use it later:
    struct timeval buttonUpTime;
    gettimeofday(&buttonUpTime, 0);

    //Paint();    // clear the canvas

    double fx = (event->pos().x() - mMoonCenterX) * mMirrored;
    double fy = event->pos().y() - mMoonCenterY;

    // Translate that to moon coordinates
    double fLatitude = -asin(fy / mMoonRadius);
    double latitude = fLatitude * 180./M_PI - mImgLibrationN;
    double longitude = asin(fx / mMoonRadius / cos(fLatitude)) * 180./M_PI
                        - mImgLibrationE;
    if (sDebugMode)
        printf("Clicked at %dE, %dN\n", (int)latitude, (int)longitude);

    // Figure out which crater lives there
    double mindist = mMoonRadius*mMoonRadius;
    MoonFeature* nearfeature = 0;
    MoonFeature* feature;
    MoonFeature::InitLoop();

    while ((feature = MoonFeature::NextFeature()) != 0)
    {
        int isInside;
        double thisdist = feature->Distance(longitude, latitude, &isInside);
        if (isInside)
        {
            // For big maria, lots of things will be inside
            // and it can be hard to click exactly on an object.
            // So for big features, don't let "inside" stop us
            // from looking for a nearby crater.
            if (feature->GetSize() < 150)
            {
                mindist = 0;
                nearfeature = 0;
            }
            mMatches[mCraterNum++] = feature;
            if (mCraterNum >= MAX_OBJECTS_SHOWN-1)
                break;
            //PaintFeature(feature);
            if (sDebugMode)
                printf("Inside %s\n", feature->GetName());
        }
        else if (thisdist < mindist)
        {
            nearfeature = feature;
            mindist = thisdist;
        }
    }

    // Draw the nearby crater if we aren't inside of one:
    if (nearfeature)
    {
        if (mCraterNum < MAX_OBJECTS_SHOWN-1)
            mMatches[mCraterNum++] = nearfeature;

        if (sDebugMode)
            printf("Landed near %s\n", nearfeature->GetName());
        PaintFeature(nearfeature);
    }

    Paint();

    // If the user held the mouse down for a while,
    // then pop up an information dialog as well:
    int dusec = 1000000 * (buttonUpTime.tv_sec - mLastButtonDownTime.tv_sec)
        + buttonUpTime.tv_usec - mLastButtonDownTime.tv_usec;
    if (dusec > 850000)
        QMessageBox::information(this, "QMoon", mDescriptions,
                                 QMessageBox::Ok);
}

